#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

# Maximaler und minimaler Wert für Begrenzung des Reglerausgangs
MAX_OUTPUT = 100
MIN_OUTPUT = 0

DEBUG = False

_logger = logging.getLogger(__name__)


class PID(object):
    """!
    PID Regler für Ofentemperatur
    """
    def __init__(self, config_manager):
        """!
        Konstruktor initialisiert den Regler und lädt die Reglerparameter
        @param config_manager: Instanz des Config Managers
        """

        # PID-Regler Parameter
        n = 1
        # Temperaturwerte für Kennfeld lesen
        self._kp_temperature_list = []
        while True:
            temp = int(config_manager.get_config_file_value('PID', 'k_p_{}_temp'.format(n)))
            if temp == 0:
                break
            else:
                self._kp_temperature_list.append(temp)
                n += 1

        # PID Werte für Kennfeld lesen
        self._kp_value_list = []
        self._ki_value_list = []
        self._kd_value_list = []
        for i in range(1, n):
            self._kp_value_list.append(float(config_manager.get_config_file_value('PID', 'k_p_{}_value'.format(i))))
            self._ki_value_list.append(float(config_manager.get_config_file_value('PID', 'k_i_{}_value'.format(i))))
            self._kd_value_list.append(float(config_manager.get_config_file_value('PID', 'k_d_{}_value'.format(i))))

        # Kennfeld erzeugen
        self._kp_value_dict = dict(zip(map(str, self._kp_temperature_list), self._kp_value_list))
        self._ki_value_dict = dict(zip(map(str, self._kp_temperature_list), self._ki_value_list))
        self._kd_value_dict = dict(zip(map(str, self._kp_temperature_list), self._kd_value_list))

        # PID Werte für Initialisierung aus Kennfeld lesen
        self._kp = self._get_best_kp_value_from_list(298)       # Annahme 25°C
        self._ki = self._get_best_ki_value_from_list(298)       # Annahme 25°C
        self._kd = self._get_best_kd_value_from_list(298)       # Annahme 25°C

        if DEBUG:
            _logger.debug("self._kp: {} ".format(self._kp))
            _logger.debug("self._ki: {} ".format(self._ki))
            _logger.debug("self._kd: {} ".format(self._kd))

        self._p = 0
        # self._p_old = 0                         # Letze Regelabweichung für D-Anteil
        self._p_list = [self._p] * 100
        self._i = 0
        self._d = 0

        # self._lastError = 0  # Letzte Regelabweichung für D-Anteil

    def clear(self):
        """!
        Setzt den Reger zurück (Werte = 0).
        """
        self._p = 0
        self._p_old = 0                         # Letze Regelabweichung für D-Anteil
        self._p_list = [self._p] * 100
        self._i = 0                             # Summe der Regelabweichung für I-Anteil
        self._d = 0

    def _get_best_kp_value_from_list(self, act_temp_value):
        """!
        Gibt den p-Faktor, dessen Temperatur-Wert am nächsten zum Parameter act_temp_value hat, zurück.
        @return p-Anteil
        """
        return self._kp_value_dict[str(min(self._kp_temperature_list, key=lambda x: abs(x - act_temp_value)))]

    def _get_best_ki_value_from_list(self, act_temp_value):
        """!
        Gibt den i-Faktor, dessen Temperatur-Wert am nächsten zum Parameter act_temp_value hat, zurück.
        @return i-Anteil
        """
        return self._ki_value_dict[str(min(self._kp_temperature_list, key=lambda x: abs(x - act_temp_value)))]

    def _get_best_kd_value_from_list(self, act_temp_value):
        """!
        Gibt den d-Faktor, dessen Temperatur-Wert am nächsten zum Parameter act_temp_value hat, zurück.
        @return d-Anteil
        """
        return self._kd_value_dict[str(min(self._kp_temperature_list, key=lambda x: abs(x - act_temp_value)))]

    def updatePIDParameters(self, temperature):
        """!
        Wechselt die PID Parameter in Abhängigkeit der angegebenen Temperatur.
        @param temperature: Temperatur, für welche der ideale PID-Parametersatz bestimmt werden soll.
        """
        new_kp_value = self._get_best_kp_value_from_list(temperature)
        if new_kp_value != self._kp:
            self._kp = new_kp_value
            if DEBUG:
                _logger.debug("change kp to: {}".format(self._kp))

        new_ki_value = self._get_best_ki_value_from_list(temperature)
        if new_ki_value != self._ki:
            self._ki = new_ki_value
            if DEBUG:
                _logger.debug("change ki to: {}".format(self._ki))

        new_kd_value = self._get_best_kd_value_from_list(temperature)
        if new_kd_value != self._kd:
            self._kd = new_kd_value
            if DEBUG:
                _logger.debug("change kd to: {}".format(self._kd))

    def update(self, t_soll, t_ist):
        """!
        Regler-Funktion muss zyklisch aufgerufen werden.
        @param t_soll: Sollwert, auf den geregelt werden soll.
        @param t_ist: Ist-Wert zur Berechnung der Regelabweichung
        @return Regler Stellwerte: [P-Regler Ergebnis, PID-Regler Ergebnis]
        """
        error = t_soll - t_ist                          # Regelabweichung / Fehler berechnen
        self._p = error * self._kp                      # P-Anteil berechnen

        # I- und D-Anteil wird nur berechnet, wenn Proportionalwert im Regelbereich liegt
        if -90 <= self._p <= 90:

            p_old = self._p
            i = 99
            while i > 90:                   # p_old = Mittelwert der letzten 10 Werte                               
                p_old += self._p_list[i]
                i -= 1
            p_old = p_old / 10

            p_akt = self._p
            i = 99
            while i > 97:                   # p_akt = Mittelwert der letzten 3 Werte
                p_akt += self._p_list[i]
                i -= 1
            p_akt = p_akt / 3

            self._d = (p_old - p_akt) * self._kd        # Differenzialanteil berechnen

            if self._p >= 5:                             # Integrator / I-Anteil
                self._i += 5 * self._ki                # Steigung begrenzen
            elif self._p <= -5:
                self._i -= 5 * self._ki
            else:
                self._i += self._p * self._ki

            if self._p >= 0:
                i_max = 100 - self._p                   # Integrator darf nur bis 100% auffüllen
                self._i = min(self._i, i_max)           # +100%
            else:
                i_max = 0
                self._i = max(self._i, i_max)           # Integrator darf nicht negativ werden

        else:
            self._i = 0
            self._d = 0

        output = self._p + self._i + self._d

        if DEBUG:
            _logger.debug("set, in, err, p, i, d, pid: {}; {}; {}; {}; {}; {}; {}".format(t_soll, t_ist, error, self._p, self._i, self._d, output))

        output = min(output, MAX_OUTPUT)        # Gesamtwert nach oben begrenzen
        output = max(output, MIN_OUTPUT)        # Gesamtwert nach unten begrenzen

        # Historie mit alten P-Anteilen aufbauen
        self._p_list.pop(0)
        self._p_list.append(round(self._p, 2))

        if False:
            _logger.debug("history: {}".format(self._p_list))

        return self._p, output
