#!/usr/bin/env python
# -*- coding: utf-8 -*-


import shutil
import zipfile
import os
import logging
import datetime
from platform import system
import base64
from time import sleep
from hashlib import md5
from threading import Timer

if system() != "Windows":
    import pyudev
else:
    import win32file

_logger = logging.getLogger(__name__)

BACKUP_ROOT_DIR = 'backup'


class USBHandler:
    """!
    Klasse erkennt, ob ein USB Gerät ein- oder ausgesteckt wurde und meldet dies über ein print.
    Stellt Funktionen zum Kopieren verschiedener Dateien zur Verfügung.
    """
    def __init__(self, config_manager, communication):
        """!
        Initialisiert die Geräte Auflistung.
        """
        _logger.debug("USBHandler init")
        self._name = "USBHandler"
        self._config_manager = config_manager
        self._communication = communication
        # serial number oven
        self._oven_serial = self._config_manager.get_production_file_value('system', 'serial-number-oven')
        # oven name
        self._oven_name = self._config_manager.get_production_file_value('system', 'oven-name')
        self._hf_directory = self._config_manager.get_parser().get('path', 'hf_directory')

        self._usbState = 'unplugged' if self._list_all_usb_devices() == [] else 'plugged'
        # size (in bytes) of blocks of data for file upload
        self.file_upload_block_size = 512
        # digits in file upload message for blocksize
        self.file_msg_block_size_digits = 3
        # digits in file upload message for block index
        self.file_msg_block_index_digits = 6
        # blocks to send for each 500ms batch
        self.gui_update_batch_size = 6
        self.gui_update_blockcount_max = 0
        self.gui_update_blockcount = 0
        self.gui_update_blocks = [] # leere liste fuer update erzeugen

        self.gui_update_in_progress = False

        _logger.debug('usbState = {}'.format(self._usbState))

    def _list_all_usb_devices(self):
        """!
        Listet alle verfügbaren USB Speicher auf.
        @return Liste mit allen verfügbaren USB Speichern
        """
        drive_list = []

        if system() != "Windows":
            context = pyudev.Context()
            for device in context.list_devices(subsystem='block'):
                if (device.get('DEVTYPE') == 'partition' and
                   device.get('ID_BUS') == 'usb'):
                    drive_list.append(self._config_manager.get_parser().get('path', 'usb'))
        else:
            drivebits = win32file.GetLogicalDrives()
            for d in range(1, 26):
                mask = 1 << d
                if drivebits & mask:
                    # here if the drive is at least there
                    drname = '%c:\\' % chr(ord('A')+d)
                    t = win32file.GetDriveType(drname)
                    if t == win32file.DRIVE_REMOVABLE:
                        drive_list.append(drname)
        return drive_list

    @staticmethod
    def _copy_files(root_src_dir, root_dst_dir):
        """!
        Kopiert alle Dateien und Verzeichnisse von root_src_dir nach root_dst_dir.
        Vorhandene Dateien werden überschrieben.
        @param root_src_dir: Quellverzeichnis
        @param root_dst_dir: Zielverzeichnis
        """
        for src_dir, dirs, files in os.walk(root_src_dir):
            dst_dir = src_dir.replace(root_src_dir, root_dst_dir, 1)
            # print 'src_dir: {}'.format(src_dir)
            # print 'dirs: {}'.format(dirs)
            # print 'files: {}'.format(files)
            # print 'dst_dir: {}'.format(dst_dir)
            if not os.path.exists(dst_dir):
                os.makedirs(dst_dir)
            for file_ in files:
                src_file = os.path.join(src_dir, file_)
                # print 'src_file: {}'.format(src_file)
                # dst_file = os.path.join(dst_dir, file_)
                # if os.path.exists(dst_file):
                #     os.remove(dst_file)
                shutil.copy(src_file, dst_dir)

    def save_user_programs(self):
        """!
        Sichert Benutzer-Programme auf den USB Stick. Es wird der erste gefundene USB Stick verwendet.
        Üblicher Weise kann nur ein USB Stick an die GS1 angesteckt werden.
        @return True, wenn der Kopiervorgang erfolgreich war.
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            root_src_dir = self._hf_directory + 'user/'
            root_dst_dir = usb_devices[0] + 'heating_files/user/'
            self._copy_files(root_src_dir, root_dst_dir)
            return True
        else:
            _logger.debug('Save programs: No usb drive found!')
            return False

    def extract_programs(self):
        """!
        Lädt Heizprogramme vom USB Stick auf die GS1.
        Diese Prozedur wird beim Aktualisieren der Firmware verwendet
        Anders als bei load_programs werden die Heizdateien in einem ZIP File
        auf dem USB Stick abgelegt
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            dst_heatfile_path = "/home/root/gs1/part_rw/heating_files/"
            root_zip_file = usb_devices[0] + "heating_files.zip"

            if os.path.exists(root_zip_file):
                dst_zip = os.path.join(dst_heatfile_path, "heating_files.zip")

                try:
                    shutil.copy(root_zip_file, dst_heatfile_path)
                except:
                    _logger.error("Could not copy heating_files.zip")
                    return False

                if self._extract_heatfiles(dst_zip):
                    os.remove(dst_zip)
                return True
            else:
                _logger.warning("Extract programs: Zip file not found")
                return False
        else:
            _logger.error("Extract programs: No usb drive found!")
            return False

    def load_programs(self):
        """!
        Lädt User-, Hersteller- und Serviceprogramme vom USB Stick auf die GS1.
        Es wird der erste gefundene USB Stick verwendet.
        Üblicher Weise kann nur ein USB Stick an die GS1 angesteckt werden.
        @return True, wenn der Kopiervorgang erfolgreich war.
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            root_src_dir = usb_devices[0] + 'heating_files/'
            root_dst_dir = self._hf_directory
            self._copy_files(root_src_dir, root_dst_dir)
            return True
        else:
            _logger.error('Load programs: No usb drive found!')
            return False

    def load_config_files(self):
        """!
        Lädt Konfigurations- und Produktionsdaten vom USB Stick. Es wird der erste gefundene USB Stick verwendet.
        Üblicher Weise kann nur ein USB Stick an die GS1 angesteckt werden.
        @return True, wenn der Kopiervorgang erfolgreich war.
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            for src_dir, dirs, files in os.walk(usb_devices[0]):
                for file_ in files:
                    if 'configfile.gs1' in file_:
                        src_file = os.path.join(src_dir, file_)
                        dst_file = self._config_manager.get_path_file_parser().get('path', 'configfile')
                    elif 'calibration_data.gs1' in file_:
                        src_file = os.path.join(src_dir, file_)
                        dst_file = self._config_manager.get_path_file_parser().get('path', 'calibration_data')
                    elif 'produktionsdaten.gs1' in file_:
                        src_file = os.path.join(src_dir, file_)
                        dst_file = self._config_manager.get_path_file_parser().get('path', 'produktionsdaten')
                    else:
                        src_file = ''
                        dst_file = ''

                    if src_file != '' and dst_file != '':
                        _logger.info('Copy {} to {}'.format(src_file, dst_file))
                        shutil.copy(src_file, dst_file)
                break  # nur root-Ordner des USB Stick durchsuchen
            return True
        else:
            _logger.error('Load config files: No usb drive found!')
            return False

    def _auto_backup(self):
        """!
        Prüft, ob das BACKUP_ROOT_DIR auf dem USB Stick vorhanden ist und erstellt dann automatisch ein Backup
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            backup_dir = usb_devices[0] + BACKUP_ROOT_DIR + '/'
            if os.path.isdir(backup_dir) and not self._communication.gui_available:
                self.backup()

    def backup(self):
        """!
        Erstellt ein Backup der Ofenkonfiguration und aller Heizprogramme (part_r und part_rw Ordner).
        @return True, wenn der Kopiervorgang erfolgreich war.
        """
        usb_devices = self._list_all_usb_devices()
        if len(usb_devices) > 0:
            try:
                # Ordnername für Backup zusammenbauen aus
                # Wurzelordner für alle Backups
                backup_dir = BACKUP_ROOT_DIR + '/'
                # Aktuellem Datum + Zeit
                backup_dir += datetime.datetime.now().strftime("%Y%m%d_%H%M") + '_'
                # Ofenname
                backup_dir += self._oven_name + '_'
                # Seriennummer
                backup_dir += self._oven_serial + '_'
                # Zusatz 'Backup'
                backup_dir += 'backup'
                # Dann alle Leerzeichen durch '_' ersetzen
                backup_dir = backup_dir.replace(' ', '_')

                # Alles aus 'part_r' sichern
                root_dst_dir = usb_devices[0] + backup_dir + '/part_r/'
                if system() == "Windows":
                    self._copy_files('../part_r', root_dst_dir)
                else:
                    self._copy_files('/home/root/gs1/part_r', root_dst_dir)

                # Alles aus 'part_rw' sichern
                root_dst_dir = usb_devices[0] + backup_dir + '/part_rw/'
                if system() == "Windows":
                    self._copy_files('../part_rw', root_dst_dir)
                else:
                    self._copy_files('/home/root/gs1/part_rw', root_dst_dir)
                return True
            except:
                _logger.error('Backup failed!')
                return False
        else:
            _logger.error('Backup: No usb drive found!')
            return False

    def start_gui_update(self):
        usb_devices = self._list_all_usb_devices()
        try:
            # USB Stick vorhanden?
            if len(usb_devices) > 0:
                # Ersten USB Stick aus Liste nach SW Update Dateien durchsuchen
                usb_path = usb_devices[0]
                _logger.info('Searching gui sw update files on device {}'.format(usb_path))
                gs1_update_file, gui_update_file = self._get_sw_update_file_name(usb_path)
                if gui_update_file == '':
                    # kein Update auf USB Stick, dann mit Meldung beenden.
                    _logger.warning('No gui sw update file found on device {} '.format(usb_devices[0]))
                    return False
                elif gui_update_file == 'build.zip':
                    if not self._do_start_gui_sw_update(usb_path):
                        _logger.warning('failed gui sw update')
                        return False
            return True
        except:
            return False

    def resend_gui_update_block(self, blockcount):
        if blockcount > self.gui_update_blockcount_max:
            return False

        block_msg = self._encode_data_blocks(blockcount, self.gui_update_blocks[blockcount])
        self._communication.write(block_msg)

        return True

    def sw_update(self, current_version):
        """!
        Startet den SW Update Vorgang. Sucht die Datei mit der neuesten Version im root-Ordner des USB Sticks.
        Wenn SW Version auf USB Stick neuer als aktuelle Version, dann wird das Update gestartet.
        @param current_version: String mit aktueller SW Version (z.B. 'R002').
        @return True, wenn der Update Vorgang erfolgreich gestartet wurde.
        """
        usb_devices = self._list_all_usb_devices()
        # USB Stick vorhanden?
        if len(usb_devices) > 0:
            # Ersten USB Stick aus Liste nach SW Update Dateien durchsuchen
            usb_path = usb_devices[0]
            _logger.info('Searching sw update files on device {}'.format(usb_path))
            gs1_update_file, gui_update_file = self._get_sw_update_file_name(usb_path)

            if gs1_update_file == '':
                # kein Update auf USB Stick, dann mit Meldung beenden.
                _logger.warning('No sw update file found on device {} '.format(usb_devices[0]))
                return False
            elif not self._do_version_check( gs1_update_file[4:-4], current_version):
                # Versionsprüfung durchgefallen
                return False
            else:
                # Versuche, neue Heizprogramme von USB Stick zu kopieren
                if self.extract_programs():
                    _logger.info("Successfully extracted heatfile programs")
                else:
                    _logger.error("Could not extract heatfile programs")

                # Entpacke SW Update von USB Stick in neuen SW Ordner
                _logger.info('Start sw update with file: {}'.format(gs1_update_file))

                new_version = gs1_update_file[4:-4]
                if not self._extract_update_file(usb_path, gs1_update_file, new_version):
                    return False
                # elif not self._do_gui_sw_update():
                    # return False
                else:
                    # Pfad in GS1 Startskript auf neuen Ordner ändern
                    _logger.debug("modify 'start_gs1.py'")
                    try:
                        # Dazu Startskript lesen
                        with open('start_gs1.py', 'r') as start_gs1:
                            start_script = start_gs1.read()
                        # Versionsnummer ersetzen
                        start_script = start_script.replace(current_version, new_version)
                        # Danach Startskript überschreiben
                        with open('start_gs1.py', 'w') as start_gs1:
                            start_gs1.write(start_script)
                    except IOError:
                        _logger.error("Can not modify 'start_gs1.py'!")
                        return False
                    else:
                        # Am Ende alles erfolgreich ausgeführt
                        _logger.info('Sw update done with file: {}'.format(gs1_update_file))
                        return True
        else:
            _logger.error('Update: No usb device found!')
            return False

    def _do_end_gui_update(self):
        self.gui_update_in_progress = False
        self._communication.set_sending_file_blocks(False)
        self._communication.enable_gui_detection(True)

        return True

    @staticmethod
    def _extract_update_file(usb_path, update_file, new_version):
        """!
        Entpackt die zip Datei mit dem SW Update.
        @param usb_path: Pfad zum USB Stick mit dem SW Update
        @param update_file: Name der SW Update Datei.
        @param new_version: Neue SW Versions Nummer.
        @return True, wenn die Datei ohne Fehler entpackt wurde.
        """
        update_file_path = usb_path + update_file
        _logger.debug('Extracting file {}'.format(update_file_path))
        try:
            with zipfile.ZipFile(update_file_path, 'r') as zip_ref:
                if system() == "Windows":
                    zip_ref.extractall('..')
                else:
                    zip_ref.extractall('.')
        except IOError:
            _logger.error('Error with update file {}'.format(update_file_path))
            return False
        else:
            _logger.info('Extracted new sw files to src_{}'.format(new_version))
            return True

    @staticmethod
    def _extract_heatfiles(heatfile_zip):
        """!
        Entpackt die zip Datei mit neuen Heatfiles.
        @param heatfile_zip: Pfad der heatfile Zip Datei.
        @return True, wenn die Datei ohne Fehler entpackt wurde.
        """
        try:
            with zipfile.ZipFile(heatfile_zip, "r") as zip_file:
                for file_path in zip_file.namelist():
                    # if a found file is of type 'hf' it will be copied from the zip file
                    # if that fails, it returns False to signal an error
                    if file_path.endswith(".hf") and not USBHandler._copy_hf_from_zip(zip_file, file_path):
                        return False
            return True

        except IOError as err:
            _logger.error("Error extracting heating_files.zip: {}".format(err))
            return False
        
    @staticmethod
    def _copy_hf_from_zip(zip_handle, path):
        dst_folder = "./part_rw/heating_files"
        if "manufacturer/" in path:
            dst_folder = os.path.join(dst_folder, "manufacturer")
        elif "user/" in path:
            dst_folder = os.path.join(dst_folder, "user")
        elif "service/" in path:
            dst_folder = os.path.join(dst_folder, "service")

        if not os.path.isdir(dst_folder):
            os.mkdir(dst_folder)

        dst_path = os.path.join(dst_folder, os.path.basename(path))

        if os.path.exists(dst_path):
            os.remove(dst_path)

        try:
            with zip_handle.open(path) as zip_hf, file(dst_path, "wb") as dst_hf:
                shutil.copyfileobj(zip_hf, dst_hf)
        except:
            _logger.error("Could not copy heatfile from zip")
            return False
        else:
            _logger.debug("Extracted heatfile {} from zip".format(os.path.basename(path)))
            return True

    @staticmethod
    def _do_version_check(new_version, current_version):
        """!
        Überprüft, ob die neue SW Version neuer ist als die vorhandene.
        Gibt eine Log-Warnung aus, wenn eine Quali oder Entwicklerversion installiert wird.
        @param new_version: SW Version die installiert werden soll
        @param current_version: Aktuell auf der GS1 installierte Version
        @return True, wenn die neue SW Version installiert werden darf.
        """
        # Warnung ausgeben, wenn eine Entwicklungs- oder Quali-Software installiert wird
        if new_version[0] != 'R':
            _logger.warning('No release version used for update: current version {}, new version {}'.format(current_version, new_version))
            return True
        # Wenn SW auf USB Stick älter ist als aktuelle Version, dann mit Meldung beenden
        elif current_version == 'R' and new_version <= current_version:
            _logger.warning('No update started: current version {}, new version {}'.format(current_version, new_version))
            return False
        # Sonst ist alles gut.
        else:
            _logger.debug('version check done')
            return True

    def _do_start_gui_sw_update(self, _usb_path):
        """!
        Liest das GUI Update File und wandelt dieses in ASCII Zeichen um.
        Sendet die ASCII Zeichen an die GUI.
        @return True, wenn der Update Vorgang erfolgreich gestartet wurde.
        """
        try:
            # Datei öffnen
            with open(_usb_path+'build.zip', 'rb') as gui_file:
                # Inhalt lesen
                #self._communication.enable_gui_detection(False)
                _logger.info('update file read')
                file_eof = False
                
                while not file_eof:
                    byte_block = gui_file.read(self.file_upload_block_size)
                    self.gui_update_blocks.append(byte_block)
                    file_eof = len(byte_block) != self.file_upload_block_size

                self.gui_update_blockcount_max = len(self.gui_update_blocks)
                self.gui_update_blockcount = 0

                # GUI vorbereiten
                self._communication.write("update=={}".format(self.gui_update_blockcount_max))
                self._communication.prepare_for_file_blocks(True)
                self._communication.enable_gui_detection(False)

                # trigger update process
                Timer(2, self.trigger_update).start()

            return True
        except IOError:
            self._communication.enable_gui_detection(True)
            _logger.error('gui update failed!')
            return False
        except:
            self._communication.enable_gui_detection(True)
            _logger.error('gui update failed - 2!')
            return False

    def trigger_update(self):
        self.gui_update_in_progress = True
        self._communication.set_sending_file_blocks(True)

    def gui_update(self):
        try:
            # return if all blocks have been sent
            if self.gui_update_blockcount >= self.gui_update_blockcount_max:
                return True

            update_batch_count = 0

            while update_batch_count < self.gui_update_batch_size:
                current_index = self.gui_update_blockcount

                # end of update file
                if current_index + 1 == self.gui_update_blockcount_max:
                    # set update batch count to max to stop loop afterwards
                    update_batch_count = self.gui_update_batch_size

                current_block = self.gui_update_blocks[current_index]
                self._communication.write(self._encode_data_blocks(current_index, current_block))
                self.gui_update_blockcount += 1
                update_batch_count += 1

            return True
        except:
            _logger.error('gui update failed!')
            return False

    def _encode_data_blocks(self, block_index, byte_block):
        data_hash = md5(byte_block).hexdigest()

        index_str = str(block_index).zfill(self.file_msg_block_index_digits)
        size_str = str(len(byte_block)).zfill(self.file_msg_block_size_digits)
        encoded_byte_block = base64.b64encode(byte_block)

        return index_str + size_str + encoded_byte_block + '&' + data_hash

    @staticmethod
    def _get_sw_update_file_name(search_path):
        """!
        Sucht SW-Update Dateien im angegebenen Pfad. Der Dateiname muss dem Muster 'GS1_*.zip' entsprechen.
        Es wird die zuletzt aufgelistete (alphabetische Sortierung) Datei verwendet.
        @param search_path: Pfad, in dem eine SW Update Datei gesucht werden soll.
        @return Name der SW Update Datei. Leerer String, wenn keine Datei gefunden wurde.
        """
        gui_update_file = ''
        gs1_update_file = ''
        for src_dir, dirs, files in os.walk(search_path):
            for file_ in files:
                # Update Datei beginnt mit 'GS1_' und endet mit '.zip'
                if ('GS1_' in file_) and ('.zip' in file_):
                    _logger.info('Found sw update file: {}'.format(file_))
                    gs1_update_file = file_
                # Update Datei 'build.zip'
                if 'build.zip' in file_:
                    _logger.info('Found gui sw update file: {}'.format(file_))
                    gui_update_file = file_
            break  # nur root-Ordner des USB Stick durchsuchen

        return gs1_update_file, gui_update_file

    def monitor(self):
        """!
        Dient der Überwachung und Erkennung, ob ein USB Stick ein-/ausgesteckt wurde. Sollte zyklisch (1s) aufgerufen werden.
        @return '' - Keine Veränderung\n
                'plugged' - USB Stick eingesteckt\n
                'unplugged' - USB Stick ausgesteckt
        """
        # Keine USB Devices gefunden und letzter Zustand ist 'eingesteckt'
        if self._list_all_usb_devices() == [] and self._usbState == 'plugged':
            # Dann wechseln auf 'ausgesteckt' und entsprechend zurückgeben
            self._usbState = 'unplugged'
            _logger.debug('usbState = unplugged')
            return self._usbState
        # USB Devices gefunden und letzter Zustand ist 'ausgesteckt'
        elif self._list_all_usb_devices() != [] and self._usbState == 'unplugged':
            # Automatische Backup initiieren
            self._auto_backup()
            # Dann wechseln auf 'eingesteckt' und entsprechend zurückgeben
            self._usbState = 'plugged'
            _logger.debug('usbState = plugged')
            return self._usbState
        # sonst keine Änderung
        else:
            # nix zurückgeben
            return ''

    def getUsbState(self):
        """!
        Gibt den aktuellen USB Status zurück.
        @return 'plugged' - USB Stick eingesteckt\n
                 'unplugged' - USB Stick ausgesteckt
        """
        return self._usbState

    def is_gui_update_in_progress(self):
        return self.gui_update_in_progress
