#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = False


class CommandParser(object):
    """!
    Parser für die Befehle von der Bedieneinheit
    """
    def __init__(self, gs1controller):
        """!
        Initialisiert den Kommando Parser.
        @param gs1controller: Instanz des GS1Controllers
        """
        _logger.debug("CommandParser init")
        self._name = "CommandParser"

        self._gs1Controller = gs1controller

    def executeCommand(self, commandstring):
        """!
        Trennt das übergebene Kommando in Befehl und Wert. Führt dieses danach aus.
        @param commandstring: Kommando und Wert als String mit '==' als Trennzeichen.
        @return Rückgabestring abhängig vom Kommando.
        """
        retstring = "nack"      # Default Rückgabewert 'nack'

        if ENABLE_DEBUG_LOG:
            _logger.debug("execute command: {}".format(commandstring))

        try:
            commandlist = commandstring.split('==', 1)
            command = commandlist[0]
        except ValueError:
            _logger.error("error when parsing command: {}".format(commandstring))
            retstring = "nack"
        else:
            if command == "ststart" or command == "redost":
                # Selbsttest starten
                retstring = "ack" if (self._gs1Controller.do_start_selftest()) else "nack"
            elif command == "hpstart":
                # Heizprogramm starten
                keys = commandlist[1].split(',')
                # self._gs1Controller.do_set_predry_phase(str(int(keys[1]) * 60))   # Vortrocknungszeit in [s] umrechnen
                retstring = "ack" if (self._gs1Controller.do_start_heat_process(keys[0], keys[1])) else "nack"
            elif command == "hpreq":
                # Heizprogramm an GUI senden
                retstring = "ack" if (self._gs1Controller.do_load_hf_data(commandlist[1])) else "nack"
            elif command == "guistate":
                # GUI Zustand zum sichern
                self._gs1Controller.set_gui_state(commandlist[1])
                retstring = "ack"
            elif command == "date":
                # Neues Datum stellen
                self._gs1Controller.set_date(commandlist[1])
                retstring = "ack"
            elif command == "time":
                # Neue Uhrzeit stellen
                self._gs1Controller.set_time(commandlist[1])
                retstring = "ack"
            elif command == "defaultsettings":
                # Werkseinstellungen
                self._gs1Controller.do_restore_default_settings()
                retstring = "ack"
            elif command == "heatfile":
                # Neue Heizdatei von der GUI zum Speichern im Ordner heating_files
                _logger.debug("received heat file data: {}".format(commandlist[1]))
                retstring = "ack" if (self._gs1Controller.do_safe_hf_data(commandlist[1])) else "nack"
            elif command == "hpchange":
                # Änderung im laufenden Heizprozess
                retstring = "ack" if (self._gs1Controller.do_change_hp_data(commandlist[1])) else "nack"
            elif command == "hpdelete":
                # Heizprogramm löschen
                retstring = "ack" if (self._gs1Controller.do_delete_hf_data(commandlist[1])) else "nack"
            elif command == "hpstop":
                # Stoppen des laufenden Heizprozesses
                self._gs1Controller.do_stop_heat_process()
                retstring = "ack"
            elif command == "tofavorite":
                # Heizprogramm als Favorit markieren
                keys = commandlist[1].split(',')
                retstring = "ack" if (self._gs1Controller.do_hf_tofavorite(keys[0], keys[1])) else "nack"
            elif command == "movedoor":
                # Türe bewegen. Entscheidung ob Öffnen oder Schließen liegt an der GS1
                retstring = "ack" if (self._gs1Controller.do_move_door()) else "nack"
            elif command == "timer":
                # Timer für Heizprogrammstart stellen
                retstring = "ack" if (self._gs1Controller.do_set_timer(commandlist[1])) else "nack"
            elif command == "update":
                # Softwareupdate starten
                retstring = "ack" if (self._gs1Controller.do_start_gui_update()) else "usb==failed"
            elif command == "gs1update":
                # Softwareupdate starten
                retstring = "ack" if (self._gs1Controller.do_update()) else "nack"
            elif command == "resend":
                # Softwareupdate starten
                retstring = "ack" if (self._gs1Controller.do_set_gui_update(commandlist[1])) else "nack"
            elif command == "endupdate":
                # Softwareupdate beenden
                retstring = "ack" if (self._gs1Controller.do_end_gui_update()) else "nack"
            elif command == "load":
                # Programme von USB Stick auf die GS1 laden
                retstring = "ack" if (self._gs1Controller.do_load_programs()) else "nack"
            elif command == "save":
                # Programme von GS1 auf USB Stick sichern
                retstring = "ack" if (self._gs1Controller.do_save_programs()) else "nack"
            elif command == "configupdate":
                # Konfiguration von USB Stick auf die GS1 laden
                retstring = "ack" if (self._gs1Controller.do_load_config_files()) else "nack"
            elif command == "backup":
                # Backup der Ofenkonfiguration erstellen
                retstring = "ack" if (self._gs1Controller.do_backup()) else "nack"
            elif command == "matchanged":
                # Bestätigung, dass Sinterschale gewechselt wurde
                self._gs1Controller.do_conf_mat_changed()
                retstring = "ack"
            elif command == "matnotchanged":
                # Sinterschale wurde nicht gewechselt
                self._gs1Controller.do_conf_mat_changed(False)
                retstring = "ack"
            elif command == "insulationinsert":
                # Bestätigung, dass Sinterschale gewechselt wurde
                self._gs1Controller.do_conf_insulationinstert()
                retstring = "ack"
            elif command == "deletefailurelist":
                # setzt auf Befehl der GUI die Fehlerliste zurueck
                self._gs1Controller.do_reset_error_memory()
                retstring = "ack"
            elif command == "lang":
                # speichert die aktuelle GUI-Sprache in operating_data.gs1
                retstring = "ack" if (self._gs1Controller.do_safe_operating_language(commandlist[1])) else "nack"
            elif command == "unit":
                # speichert die aktuelle Temperatureinheit in operating_data.gs1
                retstring = "ack" if (self._gs1Controller.do_safe_operating_unit(commandlist[1])) else "nack"
            elif command == "doorclose":
                # schließt den Türlift
                retstring = "ack" if (self._gs1Controller.do_door('close')) else "nack"
            elif command == "hfresend":
                # sendet alle heizdateien erneut
                self._gs1Controller.do_load_all_hf_data()
                retstring = "ack"
            elif command == "preheat":
                retstring = "ack"
                self._gs1Controller.set_preheat(commandlist[1])
            else:
                _logger.warning("unknown command: {}".format(commandstring))
                retstring = "nack"

        return retstring
